# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging
import psutil
import time

from threading import Thread
from datetime import datetime

log = logging.getLogger(__name__)

psutil_exceptions = (ProcessLookupError, PermissionError, psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess)


def comparetimerange(t, start_time, end_time) -> int:
    '''Returns 0 if within the time range, returns 1 if after the time range, returns -1 if before the time range'''
    def _convert_to_timestamp(t):
        if type(t) == str:
            t = float(t)
        elif type(t) == datetime:
            t = t.timestamp()
        return t

    if t is None:
        return None

    t = _convert_to_timestamp(t)

    if start_time:
        start_time = _convert_to_timestamp(start_time)
        if not start_time <= t:
            return -1
    if end_time:
        end_time = _convert_to_timestamp(end_time)
        if not t <= end_time:
            return 1

    return 0


def iswithintimerange(t, start_time, end_time):
    result = comparetimerange(t=t, start_time=start_time, end_time=end_time)
    if result is None:
        return None
    elif result == 0:
        return True
    return False


def get_subprocess_names():
    '''Return a list of subprocesses for the current pid'''
    current_process = psutil.Process()

    try:
        children_processes = current_process.children(recursive=True)
    except psutil_exceptions:
        pass

    children = []
    for child in children_processes:
        try:
            children.append(child.name())
        except psutil_exceptions:
            pass
    children.reverse()
    return children


class StatusPrinter(object):
    '''Continually prints the status until stopped.'''
    def __init__(self, timeout, max_width=None):
        self.start_time = time.time()
        self.timeout = timeout
        self.should_stop = False
        self.t = None
        self.line = ''
        self.max_width = 100
        if max_width:
            self.max_width = max_width

    def stop(self):
        self.should_stop = True
        if self.t:
            self.t.join()
        self.t = None

    def start(self):
        if not sys.stdout.isatty():
            return
        self.should_stop = False
        self.t = Thread(target=self.printer, args=())
        self.t.start()

    def printer(self):
        while not self.should_stop:
            time.sleep(.1)
            plugins = ', '.join({x for x in get_subprocess_names()})
            elapsed_time = int(time.time() - self.start_time)
            new_line = f"Time: {elapsed_time}/{self.timeout}. Running plugins: {plugins}"
            new_line = f"{new_line:.{self.max_width}}"
            if new_line == self.line:
                continue
            else:
                self.line = new_line

            # Clear line and return cursor to the beginning
            sys.stdout.write('\x1b[2K\r')  # Clear line and return cursor to the beginning
            sys.stdout.flush()

            # Write new line
            sys.stdout.write(self.line)
            sys.stdout.flush()

        # Do a final flush
        sys.stdout.write('\x1b[2K\r')
        sys.stdout.flush()


def get_traceback_variables():
    '''Intended to be called in the except step of a try/except clause to gather all the local variables for printing.'''
    exc_type, exc_value, tb = sys.exc_info()
    while True:
        if not tb.tb_next:
            break
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back
    output = []
    output.append("Printing variables by code frame, printing innermost code frame first.")
    for frame in stack:
        output.append("---Variables for frame {0} in {1} at line {2}---".format(frame.f_code.co_name, frame.f_code.co_filename, frame.f_lineno))
        for key, value in frame.f_locals.items():
            # Try / excepting because we have *no* idea what objects / variables str() is about to be called on.
            line = '{key}: {value}'
            try:
                key = str(key)
            except Exception as e:
                key = f'<ERROR PRINTING KEY, Exception: {e}>'
            try:
                value = str(value)
            except Exception as e:
                value = f'<ERROR PRINTING VALUE, Exception: {e}>'
            line = line.format(key=key, value=value)
            output.append(line)
    return '\n'.join(output)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check


class ExampleCheck(Check):

    def run(self, variables: dict):
        '''If a string returned from  this method, it will be reported as "interesting" output for this plugin'''
        if 'Debug' not in variables:
            return
        if variables['Debug'] is True:
            self.log.debug("Logging example")
            return 'This is an example of a check plugin. It only prints when Debug is True.'



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock
import pytest

from fossor.checks.netiface_errors import NetIFace


def test_parse_sar():
    """Ensures stdout is appropriately merged into the datastructure."""

    input = 'Average:         ppp0      5.00      49152.00      2064.00      64738.00      2064.00      4096.00      0.00      0.00      0.00'
    sar = NetIFace()._parse_sar(input)

    assert sar.rxerr == '5.00'
    assert sar.txerr == '2064.00'
    assert sar.rxdrop == '64738.00'
    assert sar.txdrop == '2064.00'
    assert sar.txcarr == '4096.00'


@mock.patch('fossor.plugin.Plugin.shell_call')
@mock.patch('platform.system')
@mock.patch('os.path.isdir')
def test_disk_usage(isdir_mock, system_mock, sc_mock):
    """Validate the response from for run() is correct."""
    out = 'Average:         ppp0      0.00      0.00      0.00      0.00      1.00      0.00      0.00      0.00      0.00'
    err = ''
    return_code = 0

    system_mock.return_value = 'Linux'
    isdir_mock.return_value = True
    sc_mock.return_value = out, err, return_code

    c = NetIFace().run({})
    assert c == 'stat=txdrop, interface=ppp0 has surpassed threshold. value=1.0\n'



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import re
import sys
import time
import logging

from datetime import datetime, timedelta
from fossor.checks.check import Check
from fossor.utils.misc import iswithintimerange

TIME_FORMAT = '%b %d %H:%M:%S'


class Dmesg(Check):

    def run(self, variables):
        start_time = int(variables['start_time']) if 'start_time' in variables else None
        end_time = int(variables['end_time']) if 'end_time' in variables else None

        result = self._getdmesgoutput(start_time=start_time, end_time=end_time)
        if result:
            return result

    def _getdmesgoutput(self, start_time=None, end_time=None):
        result = []
        out, err, returncode = self.shell_call('/usr/bin/dmesg', stream=True)
        boot_time = self._get_boot_time()

        dmesg_pattern = re.compile('\[(.*?)\] (.*)')

        for line in out:
            m = dmesg_pattern.match(line)
            dmesg_time = m.group(1).strip()
            timestamp = boot_time + float(dmesg_time)
            if iswithintimerange(timestamp, start_time, end_time):
                message = m.group(2)
                dt = datetime.fromtimestamp(timestamp)
                date_str = dt.strftime('%Y/%m/%d %H:%M:%S')
                result.append(' '.join([date_str, message]))

        if result:
            result.reverse()
            return '\n'.join(result)

    def _get_boot_time(self):
        with open('/proc/uptime') as f:
                uptime = float(f.read().split()[0])
        boot_time = time.time() - uptime
        return boot_time


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    variables = {}
    variables['start_time'] = (datetime.now() - timedelta(seconds=3600 * 24)).timestamp()
    variables['end_time'] = datetime.now().timestamp()
    d = Dmesg()
    result = d.run(variables)
    if result:
        print(result)
    else:
        print("No recent messages in dmesg. Add a test message like this for quick local testing: echo 'test message' | sudo tee /dev/kmsg")



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Template check class
from abc import abstractmethod
from fossor.plugin import Plugin


class Check(Plugin):

    @abstractmethod
    def run(self, variables):
        '''Variables should be a dict of variables. By default, will only report if something is outputted.'''
        pass



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.variables.variable import Variable
from fossor.reports.report import TABLE_FORMATTING_WIDTH


class TerminalWidth(Variable):
    def run(self, variables):
        '''Whatever string is returned from here will be saved as a variable'''
        out, err, return_code = self.shell_call('stty size')
        try:
            rows, columns = out.split()
            return int(columns)
        except ValueError:
            # No stty.  Most likely running headless
            return 132


class MaxPluginOutputWidth(Variable):
    def run(self, variables):
        '''If plugins keep their output width under this amount, they won't be truncated.'''
        if 'TerminalWidth' not in variables:
            return
        return int(variables['TerminalWidth']) - TABLE_FORMATTING_WIDTH



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Template Report Class
import prettytable

from fossor.utils.misc import StatusPrinter
from fossor.plugin import Plugin


TABLE_FORMATTING_WIDTH = 4  # Global so other plugins can know the width of the table formatting and truncate themselves appropriately if needed.

# TODO Add start/end times to the report


class Report(Plugin):
    def run(self, variables, report_input, **kwargs):
        '''Expects a queue holding tuples like these ('name', 'output'). Expects EOF in name or value for queue termination.'''
        pass

    def create_report_text(self, input_queue, timeout, width=None, max_lines_per_plugin=None, truncate=True, stdout=False):
        '''Default report format, can be overridden'''
        # TODO add an option to table remove formatting

        if not width:
            width = 150

        result = []
        seperator = self._create_box_seperator(width)

        def add_line(text):
            if stdout:
                print(text)
            result.append(text)

        add_line(seperator)
        add_line(self._create_box_middle(text='Report', width=width, align='c'))
        add_line(seperator)

        status_printer = StatusPrinter(max_width=width, timeout=timeout)
        while True:
            status_printer.start()
            name, output = input_queue.get()
            status_printer.stop()

            if name == 'EOF':
                break

            if truncate:
                max_width = width - TABLE_FORMATTING_WIDTH
                output = self._truncate(text=output, max_width=max_width, max_height=max_lines_per_plugin)

            # Plugin Name
            add_line(self._create_box_middle(text=f"Plugin: {name}", width=width, height=max_lines_per_plugin))

            # Plugin Output
            add_line(seperator)
            add_line(self._create_box_middle(text=output, width=width, height=max_lines_per_plugin))
            add_line(seperator)

        # Close the queue
        input_queue.close()

        return '\n'.join(result)

    def _create_box(self, text, width, align='l'):
        result = ''
        result += self._create_box_seperator(width=width) + "\n"
        result += self._create_box_middle(text=text, width=width, align=align) + "\n"
        result += self._create_box_seperator(width=width) + "\n"
        return result

    def _create_box_seperator(self, width):
        return '+' + '-' * (width - 2) + '+'

    def _create_box_middle(self, text, width=None, height=None, align='l'):
        '''Default report format, can be overridden'''

        t = prettytable.PrettyTable()
        t.hrules = prettytable.NONE
        t.header = False
        t.field_names = ['foo']
        t.align = align
        # max width must be set after field_names
        t.min_width = width
        t.max_width = width
        t.min_table_width = width
        t.max_table_width = width
        t.add_row([text])
        return t.get_string()

    def _truncate(self, text, max_width=None, max_height=None):
        if max_height:
            # Truncate line count
            text_lines = [x for x in text.splitlines()]
            original_height = len(text_lines)
            if len(text_lines) > max_height:
                text_lines = text_lines[:max_height]
            new_height = len(text_lines)
            text = '\n'.join(text_lines)
            if new_height != original_height:
                text += f'\nTruncated line count from {original_height} to {new_height}. Run with --no-truncate to stop truncation.'

        if max_width:
            text_split = text.splitlines()
            long_line_count = len([line for line in text_split if len(line) > max_width])
            # Truncate line length to fit in table
            if long_line_count > 0:
                text = '\n'.join([x[:max_width] for x in text_split])
                text += f'\nTruncated {long_line_count} lines to a width of {max_width}. Run with --no-truncate to stop truncation.'

        return text



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock

from fossor.checks.buddyinfo import BuddyInfo


@mock.patch('fossor.checks.buddyinfo.BuddyInfo._getpagesize')
@mock.patch('fossor.plugin.Plugin.shell_call')
def test_buddyinfo_report(sc_mock, pagesize_mock):
    expected = """\
No significant memory fragmentation
Smallest Page Size for system is 4096
Therefore, each columns pagesize is: 4 KiB 8 KiB 16 KiB 32 KiB 64 KiB 128 KiB 256 KiB 512 KiB 1 MiB 2 MiB 4 MiB
Output of 'cat /proc/buddyinfo' is:
Node 0, zone      DMA      2      1      1      1      1      0      1      0      1      1      3
Node 0, zone    DMA32   2592   2028   1683   1408   1060    689    423    220    115     18      0
Node 0, zone   Normal  74469  83969 174582 227090 116398  48054  23129  13454   6410      2      0"""
    out = """\
Node 0, zone      DMA      2      1      1      1      1      0      1      0      1      1      3
Node 0, zone    DMA32   2592   2028   1683   1408   1060    689    423    220    115     18      0
Node 0, zone   Normal  74469  83969 174582 227090 116398  48054  23129  13454   6410      2      0"""
    err = ''
    return_code = 0

    sc_mock.return_value = out, err, return_code
    pagesize_mock.return_value = 4096
    buddy = BuddyInfo()
    result = buddy.run({})
    assert result == expected


@mock.patch('fossor.checks.buddyinfo.BuddyInfo._getpagesize')
@mock.patch('fossor.plugin.Plugin.shell_call')
def test_buddyinfo_fragmentation(sc_mock, pagesize_mock):
    out = """\
Node 0, zone      DMA      2      1      1      1      1      0      1      0      1      1      3
Node 0, zone    DMA32   2592   2028   1683   1408   1060    689    423    220    115     18      0
Node 0, zone   Normal  74469  83969      0      0      0      0      0      0      0      0      0"""
    err = ''
    return_code = 0

    sc_mock.return_value = out, err, return_code
    pagesize_mock.return_value = 4096
    buddy = BuddyInfo()
    buddy.run({})
    assert buddy.fragmentation()


@mock.patch('fossor.plugin.Plugin.shell_call')
def test_get_page_size(sc_mock):
    out = '4096'
    err = ''
    return_code = 0

    sc_mock.return_value = out, err, return_code
    buddy = BuddyInfo()
    assert buddy._getpagesize() == 4096



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

from fossor.checks.check import Check
from fossor.engine import Fossor


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


class FailCheck(Check):
    def run(self, variables):
        foo = "bar"
        raise Exception("Testing plugin failure")
        return foo


def test_crash_verbose():
    '''Check that the variable from the above class did show up in the report'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [FailCheck]
    f.variable_plugins = set()
    f.add_variable('verbose', True)
    result = f.run(report='DictObject')

    output = result['FailCheck']

    assert 'Fail' in output
    # check that the variable foo and it's value of bar showed up in the output
    assert 'foo' in output
    assert 'bar' in output


def test_crash_not_verbose():
    '''Ensure crash report does not print when result should not be verbose.'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [FailCheck]
    f.variable_plugins = set()
    f.add_variable('verbose', False)
    result = f.run(report='DictObject')
    result.pop('Stats')
    assert len(result) == 0



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
import time
import psutil
import signal
import logging
import importlib
import setproctitle
import inspect
import pkgutil
import multiprocessing as mp

from requests.structures import CaseInsensitiveDict

import fossor.plugin
import fossor.variables
import fossor.variables.variable
import fossor.checks
import fossor.checks.check
import fossor.reports
import fossor.reports.report

from fossor.utils.misc import psutil_exceptions


class Fossor(object):
    '''
    Engine for running automated investigations using three plugin types: variables, checks, reports. The engine:
    - Collects variables
    - Runs checks using applicable variables
    - Reports noteworthy findings
    '''
    def __init__(self):
        self.log = logging.getLogger(__name__)

        self.plugin_parent_classes = [fossor.plugin.Plugin,
                                      fossor.variables.variable.Variable,
                                      fossor.checks.check.Check,
                                      fossor.reports.report.Report]

        # Variables
        self.variables = CaseInsensitiveDict()
        self.add_variable('timeout', 600)  # Default timeout

        # Imported Plugins
        self.variable_plugins = set()
        self.check_plugins = set()
        self.report_plugins = set()

        self.add_plugins()  # Adds all plugins located within the fossor module recursively

    def _import_submodules_by_module(self, module) -> set:
        """
        Import all submodules from a module.
        Returns a set of modules
        """
        if '__path__' not in dir(module):
            return set()  # No other submodules to import, this is a file
        result = set()
        for loader, name, is_pkg in pkgutil.iter_modules(path=module.__path__):
            sub_module = importlib.import_module(module.__name__ + '.' + name)
            result.add(sub_module)
            result.update(self._import_submodules_by_module(sub_module))

        return result

    def _import_submodules_by_path(self, path: str) -> set:
        """
        Import all submodules from a path.
        Returns a set of modules
        """
        # Get a list of all the python files
        # Loop over list and group modules

        path = os.path.normpath(path)
        all_file_paths = []
        for root, dirs, files in os.walk(path):
            for file in files:
                all_file_paths.append(os.path.join(root, file))

        python_file_paths = [file for file in all_file_paths if '.py' == os.path.splitext(file)[1]]

        result = set()
        for filepath in python_file_paths:
            # Create a module_name
            relative_path = filepath.split(path)[-1]
            module_name = os.path.splitext(relative_path)[0]  # Remove file extension
            module_name = module_name.lstrip(os.path.sep)  # Removing leading slash
            module_name = module_name.replace(os.path.sep, '.')  # Change / to .
            module_name = f'fossor.local.{module_name}'

            # Import with the new name
            spec = importlib.util.spec_from_file_location(module_name, filepath)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            result.add(module)

        return result

    def add_plugins(self, source=fossor):
        '''
        Recursively return a dict of plugins (classes) that inherit from the given parent_class) from within that source.
        source accepts a path as a string or a module.
        '''
        if type(source) == str:
            modules = self._import_submodules_by_path(path=source)
        else:
            modules = self._import_submodules_by_module(source)
        for module in modules:
            for obj_name, obj in module.__dict__.items():
                # Get objects from each module that look like plugins for the Check abstract class
                if not inspect.isclass(obj):  # Must check if a class otherwise the next check with issubclass will get a TypeError
                    continue
                if obj in self.plugin_parent_classes:  # If this is one of the parent classes, skip, we only want the children
                    continue
                if issubclass(obj, fossor.variables.variable.Variable):
                    self.variable_plugins.add(obj)
                elif issubclass(obj, fossor.checks.check.Check):
                    self.check_plugins.add(obj)
                elif issubclass(obj, fossor.reports.report.Report):
                    self.report_plugins.add(obj)
                else:
                    continue
                self.log.debug(f"Adding Plugin ({obj_name}, {obj})")

    def clear_plugins(self):
        self.variable_plugins = set()
        self.check_plugins = set()
        self.report_plugins = set()

    def list_plugins(self):
        variables = list(self.variable_plugins)
        checks = list(self.check_plugins)
        reports = list(self.report_plugins)
        return variables + checks + reports

    def _terminate_process_group(self, process):
        '''Kill a process group leader and its process group'''
        pid = process.pid
        pgid = os.getpgid(pid)
        if pid != pgid:
            raise ValueError(f"Trying to terminate a pid that is not a pgid leader. pid: {pid}, pgid: {pgid}.")  # This should never happen

        # Try these signals to kill the process group
        for s in (signal.SIGTERM, signal.SIGKILL):
            try:
                children = psutil.Process().children(recursive=True)
            except psutil_exceptions:
                continue  # Process is already dead

            if len(children) > 0:
                try:
                    self.log.warning(f"Issuing signal {s} to process group {pgid} because it has run for too long.")
                    os.killpg(pgid, s)
                    time.sleep(1)
                except psutil_exceptions:
                    pass
                except PermissionError:  # Occurs if process has already exited on MacOS
                    pass
        # Terminate the plugin process (which is likely dead at this point)
        process.terminate()
        process.join()

    def _run_plugins_parallel(self, plugins):
        '''
        Accepts function to run, and a dictionary of plugins to run. Format is name:module
        Returns an output queue which outputs two values: plugin-name, output. Queue ends when plugin-name is EOF.

        '''

        def _run_plugins_parallel_helper(timeout):
            setproctitle.setproctitle('Plugin-Runner')
            processes = []
            queue_lock = mp.Lock()

            for Plugin in plugins:
                process = mp.Process(target=self.run_plugin, name=Plugin.get_name(), args=(Plugin, output_queue, queue_lock))
                process.start()
                processes.append(process)

            start_time = time.time()
            should_end = False
            while len(processes) > 0:

                time_spent = time.time() - start_time
                should_end = time_spent > timeout

                try:
                    os.kill(os.getppid(), 0)  # Check if the parent process is alive
                except ProcessLookupError:
                    should_end = True

                for process in processes:  # Using list so I can delete while iterating
                    process_is_dead = not process.is_alive() and process.exitcode is not None
                    if process_is_dead:
                        process.join()
                        processes.remove(process)
                        continue
                    if should_end:
                        with queue_lock:
                            self.log.error(f"Process {process} for plugin {process.name} ran longer than the timeout period: {timeout}")
                            if self.variables.get('verbose'):
                                output_queue.put((process.name, 'Timed out (use --time-out to increase timeout)'))
                            self._terminate_process_group(process)
                            processes.remove(process)
                time.sleep(.1)

            with queue_lock:
                output_queue.put(('Stats', f'Ran {len(plugins)} plugins.'))
                output_queue.put(('EOF', 'EOF'))  # Indicate the queue is finished

            return output_queue

        timeout = int(self.variables.get('timeout'))

        # Run plugins in parallel
        # This child process will spawn child processes for each plugin, and then do the final termination and clean-up.
        # A separate child process for spawning plugin processes is needed because only a parent process can terminate its children.
        # This allows us to continue to the reporting phase while the checks finish.
        output_queue = mp.Queue()
        parallel_plugins_helper_process = mp.Process(target=_run_plugins_parallel_helper, name='parallel_plugins_helper_process', args=(timeout, ))
        parallel_plugins_helper_process.start()

        return output_queue

    def run_plugin(self, Plugin, output_queue, queue_lock):
        try:
            os.setsid()  # Causes this plugin to become a process group leader, we can then kill that process group to kill all potential subprocesses
            p = Plugin()
            setproctitle.setproctitle(p.get_name())
            output = ''
            if p.should_run():
                output = p.run_helper(variables=self.variables)
                if output:
                    with queue_lock:  # Prevents the queue from becoming corrupt on process termination
                        output_queue.put((p.get_name(), output))
        except Exception as e:
            # This ensures that any plugins that fail to initialize show up properly in the report and don't output errors mid report.
            self.log.exception(f"Plugin {p} failed to initialize", e)

    def get_variables(self):
        '''
        Gather all possible variables that may be useful for plugins
            - Variables can have interdependencies, this will continue gathering missing variables
              until no new variables have appeared after a run
            - Can be run again if new variables are added to fill in any blanks
        '''
        while True:
            variable_count = len(self.variables)
            output_queue = self._run_plugins_parallel(self.variable_plugins)
            while True:
                name, output = output_queue.get()
                if 'EOF' in name:
                    break
                if name in self.variables:
                    self.log.debug("Skipping variable plugin because it has already been found: {vp}".format(vp=name))
                    continue
                self.add_variable(name, output)
            output_queue.close()

            if variable_count == len(self.variables):
                self.log.debug("Done finding variables: {variables}".format(variables=self.variables))
                break

    def _convert_simple_type(self, value):
        '''Convert String to simple type if possible'''
        if type(value) == str:
            if 'false' == value.lower():
                return False
            if 'true' == value.lower():
                return True
            try:
                return int(value)
            except ValueError:
                pass
            try:
                return float(value)
            except ValueError:
                pass

        return value

    def add_variable(self, name, value):
        '''Adds variable, converts numbers and booleans to their types'''
        if name in self.variables:
            old_value = self.variables[name]
            if value != old_value:
                self.log.warning(f"Variable {name} with value of {old_value} is being replaced with {value}")

        value = self._convert_simple_type(value)

        self.variables[name] = value
        self.log.debug("Added variable {name} with value of {value} (type={type})".format(name=name, value=value, type=type(value)))
        return True

    def run(self, report='StdOut', check_whitelist=[], check_blacklist=[], variable_plugin_whitelist=[], variable_plugin_blacklist=[]):
        '''Runs Fossor with the given report. Method returns a string of the report output.'''

        # Gather Variables
        if variable_plugin_whitelist:
            self.variable_plugins = {plugin for plugin in self.variable_plugins if plugin.get_name() in variable_plugin_whitelist}
        for variable_plugin in variable_plugin_blacklist:
            self.variable_plugins = {plugin for plugin in self.variable_plugins if plugin.get_name() not in variable_plugin_blacklist}
        self.get_variables()

        # Run Checks
        if check_whitelist:
            self.check_plugins = {plugin for plugin in self.check_plugins if plugin.get_name() in check_whitelist}
        if check_blacklist:
            self.check_plugins = {plugin for plugin in self.check_plugins if plugin.get_name() not in check_blacklist}
        output_queue = self._run_plugins_parallel(self.check_plugins)

        # Run Report
        try:
            Report_Plugin = [plugin for plugin in self.report_plugins if report.lower() == plugin.get_name().lower()].pop()
        except IndexError:
            message = f"Report_Plugin: {report}, was not found. Possibly reports are: {[plugin.get_name() for plugin in self.report_plugins]}"
            self.log.critical(message)
            raise ValueError(message)

        report_plugin = Report_Plugin()
        return report_plugin.run(variables=self.variables, report_input=output_queue)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Returns a json string object.

import json

from fossor.reports.report import Report
from fossor.reports.dict_object_report import DictObject


class Json(Report):
    def run(self, variables, report_input, **kwargs):
        do = DictObject()
        dict_report = do.run(variables, report_input, **kwargs)
        json_str = json.dumps(dict_report, sort_keys=True, indent=4)
        print(json_str)
        return json_str



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import time
import logging
import unittest

from datetime import datetime
from unittest.mock import patch
from fossor.checks.dmesg import Dmesg
from fossor.utils.misc import iswithintimerange
from fossor.utils.misc import comparetimerange

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestTimeWithinRange(unittest.TestCase):

    def test_is_none(self):
        t = time.time()
        start = t - 5
        end = t + 5
        assert comparetimerange(t=None, start_time=start, end_time=end) is None
        assert iswithintimerange(t=None, start_time=start, end_time=end) is None

        # Same test with datetimes
        t = datetime.fromtimestamp(t)
        start = datetime.fromtimestamp(start)
        end = datetime.fromtimestamp(end)
        assert iswithintimerange(t=None, start_time=start, end_time=end) is None

    def test_within_range(self):
        t = time.time()
        start = t - 5
        end = t + 5
        assert iswithintimerange(t=t, start_time=start, end_time=end) is True

        # Same test with datetimes
        t = datetime.fromtimestamp(t)
        start = datetime.fromtimestamp(start)
        end = datetime.fromtimestamp(end)
        assert iswithintimerange(t=t, start_time=start, end_time=end) is True

    def test_before_range(self):
        # Test before range
        t = time.time()
        start = t + 5
        end = t + 10
        assert iswithintimerange(t=t, start_time=start, end_time=end) is False

        # Same test with datetimes
        t = datetime.fromtimestamp(t)
        start = datetime.fromtimestamp(start)
        end = datetime.fromtimestamp(end)
        assert iswithintimerange(t=t, start_time=start, end_time=end) is False

    def test_after_range(self):
        t = time.time()
        start = t - 10
        end = t - 5
        assert iswithintimerange(t=t, start_time=start, end_time=end) is False

        # Same test with datetimes
        t = datetime.fromtimestamp(t)
        start = datetime.fromtimestamp(start)
        end = datetime.fromtimestamp(end)
        assert iswithintimerange(t=t, start_time=start, end_time=end) is False


def test_get_variables():
    '''Confirm we get the correct variables back if an exception occurs.'''
    with patch('fossor.plugin.Plugin.shell_call') as patched_shell_call:
        patched_shell_call.side_effect = Exception('foobar')
        d = Dmesg()
        variables = {}
        variables['verbose'] = True
        result = d.run_helper(variables)
        assert "args: ('/usr/bin/dmesg',)" in result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging
from fossor.checks.dmesg import Dmesg
from io import StringIO
from unittest.mock import patch

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)

example_dmesg_output = '''[2047964.712926] usb 2-1.3.4: New USB device strings: Mfr=1, Product=2, SerialNumber=3
[2047964.712929] usb 2-1.3.4: Product: Pixel
[2047964.712934] usb 2-1.3.4: SerialNumber: FA6CM0300679
[2052025.266882] usb 2-1.3.4: USB disconnect, device number 39
[2062321.522490] usb 2-1.3.4: new high-speed USB device number 40 using ehci-pci
[2062321.610727] usb 2-1.3.4: New USB device found, idVendor=18d1, idProduct=4ee1
[2062321.610732] usb 2-1.3.4: New USB device strings: Mfr=1, Product=2, SerialNumber=3
[2062321.610735] usb 2-1.3.4: Product: Pixel
[2062321.610739] usb 2-1.3.4: SerialNumber: FA6CM030067'''


def test_dmesg():
    variables = {}
    d = Dmesg()
    with patch('fossor.plugin.Plugin.shell_call') as mocked_shell_call:
        text = StringIO(initial_value=example_dmesg_output)
        mocked_shell_call.return_value = (text, '', 0)
        result = d.run(variables)
        assert 'high-speed' in result


def test_dmesg_with_start_end_times():
    variables = {}
    variables['start_time'] = 1509647158.0
    variables['end_time'] = 1509647160.0
    d = Dmesg()
    with patch('fossor.plugin.Plugin.shell_call') as mocked_shell_call:
        with patch('fossor.checks.dmesg.Dmesg._get_boot_time') as mocked_get_boot_time:
            mocked_get_boot_time.return_value = 1509664739.8043845 - 2069605.89
            text = StringIO(initial_value=example_dmesg_output)
            mocked_shell_call.return_value = (text, '', 0)
            result = d.run(variables)
            assert 'USB disconnect, device number 39' in result
            assert len(result.splitlines()) == 1



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.variables.variable import Variable
from getpass import getuser


class OtherUsers(Variable):
    """Checks for other users logged into the host box other than the current user."""
    def run(self, variables):
        # Users that we consider "uninteresting" so we won't return them.
        boring_users = [getuser(), 'root', 'app']
        out, err, return_code = self.shell_call('users')
        other_users = sorted(set(out.split()) - set(boring_users))

        if len(set(out.split()) - set(boring_users)) > 0:
            return ' '.join(str(u) for u in other_users)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.other_users import OtherUsers


def test_check_other_users():
    variables = {}
    variables['OtherUsers'] = 'phfry brodriguez tleela'
    c = OtherUsers()

    # Test for multiple users logged in
    assert c.run(variables) == 'Other users logged into this box: phfry, brodriguez, tleela\n'



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

from fossor.engine import Fossor

import fossor.variables.hostname
import fossor.checks.buddyinfo

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_convert_simple_type():
    f = Fossor()
    assert f._convert_simple_type('false') is False
    assert f._convert_simple_type('False') is False
    assert f._convert_simple_type(False) is False

    assert f._convert_simple_type('true') is True
    assert f._convert_simple_type('True') is True
    assert f._convert_simple_type(True) is True

    assert f._convert_simple_type('1.0') == 1.0
    assert type(f._convert_simple_type('1.0')) == float

    assert f._convert_simple_type('1') == 1
    assert type(f._convert_simple_type('1')) == int

    assert f._convert_simple_type('foo') == 'foo'
    assert type(f._convert_simple_type('foo')) == str


def test_variable_plugin_whitelist():
    f = Fossor()
    f.add_variable('timeout', 1)
    assert len(f.variable_plugins) > 2
    whitelist = ['Hostname', 'PidExe']
    f.run(variable_plugin_whitelist=whitelist)
    assert len(f.variable_plugins) == 2


def test_variable_plugin_blacklist():
    f = Fossor()
    f.add_variable('timeout', 1)
    assert fossor.variables.hostname.Hostname in f.variable_plugins
    blacklist = ['Hostname']
    f.run(variable_plugin_blacklist=blacklist)
    assert 'Hostname' not in f.variable_plugins
    assert len(f.variable_plugins) > 0


def test_check_plugin_whitelist():
    f = Fossor()
    f.add_variable('timeout', 1)
    assert len(f.check_plugins) > 2
    whitelist = ['BuddyInfo', 'LoadAvg']
    f.run(check_whitelist=whitelist)
    assert len(f.check_plugins) == 2
    assert fossor.checks.buddyinfo.BuddyInfo in f.check_plugins


def test_check_plugin_blacklist():
    f = Fossor()
    f.add_variable('timeout', 1)
    assert len(f.check_plugins) > 2
    assert fossor.checks.buddyinfo.BuddyInfo in f.check_plugins
    blacklist = ['BuddyInfo', 'LoadAvg']
    f.run(check_blacklist=blacklist)
    assert fossor.checks.buddyinfo.BuddyInfo not in f.check_plugins
    assert len(f.check_plugins) > 1



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
import gzip
import logging

from datetime import datetime

from fossor.utils.misc import comparetimerange, iswithintimerange


class FileTools(object):

    def __init__(self, date_format=None):
        self.log = logging.getLogger(__name__)

        self.date_format = '%Y-%m-%dT%H:%M:%S.%f'  # default date format - ISO 8601.
        if date_format:
            self.date_format = date_format
        self.date_length = None

    def get_first_last_lines(self, file_handle):
        original_position = file_handle.tell()

        file_handle.seek(os.SEEK_SET)
        first = file_handle.readline()

        file_handle.seek(0, os.SEEK_END)
        last = self._get_previous_line(file_handle=file_handle)
        file_handle.seek(original_position)
        return first, last

    def _gettimestamp(self, line) -> float:
        '''Uses the FileTools.date_format object to pull a timestamp from a line
           Will attempt to trim line to date using FileTools.date_length'''
        date_str = line[:self.date_length] if self.date_length else line
        try:
            dt = datetime.strptime(date_str, self.date_format)
            if dt.year == 1900:  # Year was missing and defaulted to 1900
                dt = dt.replace(year=datetime.now().year)
            return dt.timestamp()
        except ValueError as e:
            error_message = e.args[0]
            if 'does not match format' in error_message:
                return None
            elif 'unconverted data remains: ' in error_message:
                unconverted_data = error_message.split('unconverted data remains: ')[1]
                if not self.date_length:
                    self.date_length = len(line) - len(unconverted_data)
                    return self._gettimestamp(line)  # Try again one time now that we have a length for the date_str
                else:
                    raise e

    def _open_log_file(self, file_path):
        '''Determine if file is gzipped or not, and then return an appropriate file handle.'''
        try:
            f = gzip.open(file_path, 'rt')
            f.readline()  # Try reading a line to check if this is a gzipped file
            f.seek(0, os.SEEK_SET)  # Return seek position to beginning
            return f
        except (IOError, OSError):
            f = open(file_path, 'rt')
            return f

    def _get_previous_line(self, file_handle) -> str:
        '''Returns the previous line, even if the current seek position is in the middle of a line.
           Seek position will be placed at the first character of the returned line.
           If at the beginning of a file, the first line will be returned.'''
        f = file_handle
        self._get_current_line(file_handle=f)
        new_position = max(0, f.tell() - 2)  # Go to 1 past the current newline
        f.seek(new_position)
        return self._get_current_line(file_handle=f)

    def _get_current_line(self, file_handle) -> str:
        '''Returns the current line by seeking leftwards to the first newline or beginning of file.
           Seek position will be placed at first character of the returned line.'''
        f = file_handle
        orig_pos = f.tell()
        f = file_handle
        curr_pos = orig_pos
        while curr_pos > 0:
            curr_pos -= 1
            f.seek(curr_pos)
            try:
                char = f.read(1)
                if '\n' in char:
                    curr_pos = f.tell()  # Reset the curr_pos to just after the newline
                    break
            except UnicodeDecodeError:  # Incase we read the middle of a unicode character instead of the beginning of it
                pass  # Next pass will move one byte to the left and try to read again
        f.seek(curr_pos)
        line = f.readline()
        f.seek(curr_pos)
        return line

    def _get_first_log_line_position_binary_search(self, file_handle, start_time, end_time) -> int:
        '''Return seek position of first log line in the time range specified.'''
        f = file_handle
        eof = f.seek(0, os.SEEK_END)
        end = eof
        start = f.seek(0, os.SEEK_SET)
        earliest_pos_in_range = None
        while True:
            curr_pos = int((start + end) / 2)
            f.seek(curr_pos)

            if start == end:
                break  # We're at the start or end of the file

            # Get to the beginning of the line
            line = self._get_current_line(f)
            dateless_lines_allowed = 200
            for lines_without_dates in range(dateless_lines_allowed + 1):
                line = f.readline()
                line_comparison = comparetimerange(self._gettimestamp(line), start_time=start_time, end_time=end_time)

                if line_comparison is not None or not f.tell() < eof:
                    self._get_previous_line(f)  # reset cursor back to beginning of the line we got from readline
                    line_start = f.tell()
                    break

            if lines_without_dates >= dateless_lines_allowed:
                raise ValueError(f"Couldn't find line with valid time format in {dateless_lines_allowed} lines")
            elif line_comparison is None:  # File does not contain lines within time range
                break
            elif line_comparison == 0:  # Current line is within the time range
                earliest_pos_in_range = line_start
                end = curr_pos
            elif line_comparison < 0:  # Current line is earlier than the time range
                start = min(curr_pos + 1, end)
            elif line_comparison > 0:  # Current line is later than the time range
                end = max(curr_pos - 1, start)

        timestamp = self._gettimestamp(line)
        self.log.debug(f"Earliest line in range {start_time}-{end_time} at position: {earliest_pos_in_range} with timestamp: {timestamp}. Line: {repr(line)}")
        return earliest_pos_in_range

    def get_logs_in_time_range(self, file_paths, start_time=None, end_time=None):
        timestamped_lines = self.get_logs_in_time_range_with_timestamps(file_paths=file_paths, start_time=start_time, end_time=end_time)
        return (line for timestamp, line in timestamped_lines)

    def get_logs_in_time_range_with_timestamps(self, file_paths, start_time=None, end_time=None):
        '''Return a generator that provides the log lines within the specified times.
           Orders files based on timestamp present in first log line if available.
           Lines not matching FileTools.date_format will not be returned.'''

        file_handles = []
        for file_path in file_paths:
            f = self._open_log_file(file_path)
            first, last = self.get_first_last_lines(f)

            first_timestamp = self._gettimestamp(first)
            last_timestamp = self._gettimestamp(last)

            # Skip the file if the timestamps aren't in the right range
            if not iswithintimerange(first_timestamp, start_time=None, end_time=end_time):
                self.log.debug(f"Skipping file {file_path} since first timestamp in file {first_timestamp} is after the end_time: {end_time}")
                continue
            if not iswithintimerange(last_timestamp, start_time=start_time, end_time=None):
                self.log.debug(f"Skipping file {file_path} since last timestamp in file {first_timestamp} is before the start_time: {start_time}")
                continue

            file_handles.append((f, first_timestamp))

        file_handles.sort(key=lambda x: x[1])
        file_handles = [f for f, timestamp in file_handles]  # Remove the timestamps so file_handles is storing sorted file_handles without timestamps now.

        for f in file_handles:
            pos = self._get_first_log_line_position_binary_search(file_handle=f, start_time=start_time, end_time=end_time)
            f.seek(pos)
            for line in f:
                timestamp = self._gettimestamp(line)
                if iswithintimerange(timestamp, start_time=start_time, end_time=end_time):
                    yield timestamp, line
                else:
                    continue



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock

from fossor.checks.thcount import Thcount


@mock.patch('fossor.plugin.Plugin.shell_call')
def test_thcount(sc_mock):
    out = 'THCNT USER\n1 root\n2 ermanuel\n1001 scallist\n10000 fossor \n1 fossor\n1 root\n1 root\n1 root\n1 root\n1 root\n1 root'
    err = ''
    return_code = 0

    sc_mock.return_value = out, err, return_code
    c = Thcount()
    assert c.run({}) == 'user fossor currently has 10001 threads\n'



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock

from fossor.checks.diskusage import DiskUsage


@mock.patch('fossor.plugin.Plugin.shell_call')
def test_disk_usage(sc_mock):
    out = 'Filesystem     1K-blocks      Used Available Use% Mounted on\n' \
          '/dev/sda2       407G  318G   69G  99% /\n' \
          'tmpfs            32G  161M   32G   1% /dev/shm'
    err = ''
    return_code = 0

    sc_mock.return_value = out, err, return_code
    c = DiskUsage()
    assert c.run({}) == 'Disk utilization is at critical state (> 98). partition=/dev/sda2 at utilization=99%'



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.variables.variable import Variable
import socket


class Hostname(Variable):
    def run(self, variables):
        return socket.getfqdn()



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

from fossor.cli import main
from click.testing import CliRunner

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_fossor_help_menu():
    runner = CliRunner()
    result = runner.invoke(main, args=['-h', '--timeout 1'])
    assert result.exit_code == 0
    assert 'Usage' in result.output
    assert '--help' in result.output


def test_fossor_extra_kwarg():
    runner = CliRunner()
    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', 'foo=2', '--verbose', '--no-truncate'])
    assert result.exit_code == 0
    assert 'foobar=5' in result.output
    assert 'foo=2' in result.output


def test_fossor_invalid_kwarg():
    runner = CliRunner()
    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', '--not-a-valid-flag'])
    assert result.exit_code == -1
    assert '' == result.output

    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', '--another-invalid-flag'])
    assert result.exit_code == -1
    assert '' == result.output

    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', '6'])
    assert result.exit_code == -1
    assert '' == result.output

    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', '--another-invalid-flag', '6'])
    assert result.exit_code == -1
    assert '' == result.output

    result = runner.invoke(main, args=['--time-out', '1', 'foobar=5', '--not-a-valid-flag', '--another-invalid-flag', '6'])
    assert result.exit_code == -1
    assert '' == result.output



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import random
import logging

from fossor.utils import anomaly_detection

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_rare_distribution():
    foo = [random.randint(0, 10) for x in range(1000)]
    assert not anomaly_detection.abnormal_distribution(foo)

    foo = [random.randint(0, 10) for x in range(1000)] + [11 for x in range(25)]
    assert anomaly_detection.abnormal_distribution(foo)

    foo = [random.randint(0, 10) for x in range(500)] + [5 for x in range(100)] + [random.randint(0, 10) for x in range(500)]
    assert anomaly_detection.abnormal_distribution(foo)

    foo = [random.randint(0, 10) for x in range(1000)] + [14 for x in range(15)]
    assert anomaly_detection.abnormal_distribution(foo)


def test_within_three_sigma():
    values = [1 for x in range(1000)]
    assert anomaly_detection.within_three_sigma(values=values)


def test_outside_two_and_three_sigma():
    values = [1 for x in range(1000)] + [1000 for x in range(100)]
    assert not anomaly_detection.within_three_sigma(values=values)


def test_outside_two_and_three_2_sigma():
    values = [1 for x in range(100)] + [50 for x in range(1000)] + [1000 for x in range(100)]
    assert not anomaly_detection.within_three_sigma(values=values)


def test_outside_third_sigma():
    values = [1 for x in range(50)] + [50 for x in range(1000)] + [1000 for x in range(50)]
    assert not anomaly_detection.within_three_sigma(values=values)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check
from collections import defaultdict


class Thcount(Check):
    '''Simple thread count via ps.  Arbitrarily set to 10k threads'''
    def run(self, variables):
        tc = defaultdict(int)
        rd = ""
        out, err, return_code = self.shell_call('ps -e -o thcount,user')
        ps = out.splitlines()
        for pid in ps:
            count, user = pid.split()
            if count.isdigit():
                tc[user] += int(count)

        for user in tc:
            if tc[user] > 10000:
                rd = rd + 'user {0} currently has {1} threads\n'.format(user, tc[user])
                return rd


if __name__ == '__main__':
    t = Thcount()
    print(t.run({}))



#!/usr/bin/env python
# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.


from __future__ import absolute_import
from __future__ import print_function

import sys
import click
import logging
import setproctitle
import parsedatetime as pdt
from datetime import datetime, timedelta
from fossor.engine import Fossor

default_plugin_dir = '/opt/fossor'


@click.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True, help_option_names=['-h', '--help']))  # noqa: C901
@click.pass_context
@click.option('-p', '--pid', help='Pid to investigate.')
@click.option('--product', help='Product to investigate.')
@click.option('--plugin-dir', default=default_plugin_dir, help=f'Import all plugins from this directory. Default dir: {default_plugin_dir}')
@click.option('--hours', default=1, help='Sets start-time to X hours ago, defaults to 24 hours. Plugins may optionally implement and use this.')
@click.option('-r', '--report', default='StdOut', help='Report Plugin to run.')
@click.option('--start-time', default='', help='Plugins may optionally implement and use this.')
@click.option('--end-time', default='', help='Plugins may optionally implement and use this. Defaults to now.')
@click.option('-t', '--time-out', default=600, help='Default timeout for plugins.')
@click.option('-d', '--debug', is_flag=True)
@click.option('-v', '--verbose', is_flag=True)
@click.option('--no-truncate', is_flag=True)
def main(context, pid, product, plugin_dir, hours, report, start_time, end_time, time_out, debug, verbose, no_truncate):
    """Fossor is a plugin oriented tool for automating the investigation of broken hosts and services.

    \b
    Advanced:
    Multiple additional internal variables may be passed on the command line in this format: name="value".
    This is intended for testing or automation.
    """  # \b makes click library honor paragraphs
    setproctitle.setproctitle('fossor ' + ' '.join(sys.argv[1:]))
    if debug:
        log_level = logging.DEBUG
        verbose = True
    else:
        log_level = logging.CRITICAL
        logging.getLogger("requests").setLevel(logging.WARNING)
    logging.basicConfig(stream=sys.stdout, level=log_level)
    log = logging.getLogger(__name__)
    f = Fossor()

    # Add pre-defined args
    f.add_variable('timeout', time_out)
    if product:
        f.add_variable('product', product)
    if pid:
        f.add_variable('pid', pid)
    if no_truncate:
        f.add_variable('truncate', False)
    f.add_variable('verbose', verbose)

    # Add dynamic args
    log.debug(f"Remaining Arguments that will be used for dynamic args now that hardcoded cli flags have been removed: {context.args}")
    format_help_message = "Please use --help to see valid flags. Or use the name=value method for setting internal variables."
    for arg in context.args:
        a = arg.strip()
        if a.startswith('-'):
            raise ValueError(f"Argument: {arg} is invalid {format_help_message}.")
        try:
            name, value = a.split('=')
            f.add_variable(name, value)
            log.debug(f"Using dynamic variable {name}={value}")
        except Exception as e:
            log.exception(f"Failed to add argument: \"{arg}\". {format_help_message} Exception was: {e}")
            raise e

    # Get start/end times
    cal = pdt.Calendar()
    if start_time:
        start_time = (cal.parseDT(start_time)[0]).timestamp()
    else:
        start_time = (datetime.now() - timedelta(hours=hours)).timestamp()
    f.add_variable('start_time', str(start_time))
    if end_time:
        end_time = (cal.parseDT(end_time)[0]).timestamp()
    else:
        end_time = datetime.now().timestamp()
    f.add_variable('end_time', str(end_time))

    # Import plugin dir if it exists
    f.add_plugins(plugin_dir)

    log.debug("Starting fossor")
    return f.run(report=report)


if __name__ == '__main__':
    main()



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
import subprocess
import io
import traceback

from fossor.utils.misc import get_traceback_variables
from abc import ABCMeta, abstractmethod


class Plugin(metaclass=ABCMeta):
    '''Super class for Variable and Check plugins'''

    def should_run(self):
        '''By default always run'''
        return True

    @classmethod
    def get_name(cls):
        return cls.__name__

    @classmethod
    def get_full_name(cls):
        return cls.__module__ + '.' + cls.get_name()

    @property
    def log(self):
        return logging.getLogger(self.get_full_name())

    def shell_call(self, cmd, stream=False):
        '''
        Cmd accept a string to run in the shell
        return_code=False

        Output:
        out (string), err (string), returncode (int)

        Output if stream=True:
        out (string stream), err (string stream), returncode (int)
        '''
        if type(cmd) == str:
            cmd = cmd.split()
        self.log.debug(f"Executing shell command: {cmd}")
        p = subprocess.Popen(cmd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             close_fds=True,
                             shell=False,
                             encoding='utf-8',
                             start_new_session=False)
        if stream:
            self.log.debug(f"Not printing output since command being streamed: {cmd}")
            return p.stdout, p.stderr, p.returncode
        out, err = p.communicate()
        self.log.debug(f"Command: {cmd}, had a stdout of: {out}")
        if err:
            self.log.warning(f"Command: {cmd}, had a stderr of: {err}")
        return out, err, p.returncode

    @abstractmethod
    def run(self, variables: dict):
        '''By default, this will only report if something is outputted.'''
        pass

    def run_helper(self, variables: dict):
        '''Helper method, tracks output for plugins that want to use run as their only method. Returns a string as output.'''
        self.error = None
        self.output = None
        self.log.debug("Starting")
        try:
            self.output = self.run(variables)
            self.log.debug("Finished")
        except Exception as e:
            with io.StringIO() as output:
                traceback.print_tb(e.__traceback__, file=output)
                output.seek(0)
                trace = output.read()
            trace_variables = get_traceback_variables()
            self.error = f"Crash Report (Execution Failed)\n---exception---\n{e}\n---traceback---\n{trace}\n---variables---\n{trace_variables}"
        if self.should_notify():
            return self.output
        if not variables.get('verbose'):
            return
        if self.error:
            self.log.debug(f"Error Report: {self.error}")
            return self.error
        if self.output:
            return self.output

    def should_notify(self):
        '''Return True if the Plugin output should be returned. Can be overridden by a subclass.'''
        if self.output:
            return True



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import time
import re
import platform
from fossor.checks.check import Check


class MemUsage(Check):
    """This Check inspects the current memory usage and will alert the
    user if it seems excessive"""

    def run(self, variables):
        os_name = platform.system()
        if os_name != "Linux":
            return

        # Critical memory-in-use threshold %
        CRITICAL_THRESH = 90
        meminfo = self.get_meminfo()

        act_free = meminfo['MemFree'] + meminfo['Buffers'] + meminfo['Cached']
        used = meminfo['MemTotal'] - act_free
        perc_used = (used/meminfo['MemTotal'])*100

        out, err, return_code = self.shell_call('dmesg')

        if perc_used >= CRITICAL_THRESH or 'oom-killer' in out:
            res = 'High Memory Use!\nMemTotal: {0:.0f} kib\nUsed: {1:.0f} kib ({2:.0f}%)'.format(
                meminfo['MemTotal'], used, perc_used)

            if 'oom-killer' in out:
                res = res + '\n\noom-killer present in dmesg | tail!\n{}'.format(self.get_ooms(out))

            return res

    def get_meminfo(self):
        meminfo = {}
        with open('/proc/meminfo') as m:
            for i in m.readlines():
                meminfo[i.split()[0].rstrip(':')] = int(i.split()[1])

        return meminfo

    def get_time(self):
        now = time.time()
        up = float(open('/proc/uptime').read().split()[0])
        booted = now - up
        cutoff = up - 86400
        return booted, cutoff

    def get_ooms(self, out):
        booted, cutoff = self.get_time()
        newline = ""

        for line in out.split('\n'):
            if 'oom-killer' in line:
                t = re.match("^\[(\d+.\d+)\]", line)
                if t:
                    human_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.gmtime(booted + float(t.group(1))))
                    if float(t.group(1)) > cutoff:
                        newline = newline + human_time + line + "\n"

                else:
                    return '\n'.join([line for line in out.split('\n') in line])

        return newline


if __name__ == '__main__':
    m = MemUsage()
    print(m.run({}))



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging
from multiprocessing import Queue

from fossor.reports.report import Report


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_create_box_seperator():
    r = Report()
    for width in range(10, 21):
        result = r._create_box_seperator(width=width)
        assert '+' in result
        assert '-' in result
        assert len(result) == width


def test_create_box_middle():
    r = Report()

    for text in ['foo', 'foo' * 100]:
        for width in range(20, 41):
            result = r._create_box_middle(text=text, width=width)
            result = result
            assert 'foo' in result
            assert len(max(result.splitlines())) == width


def test_create_box():
    r = Report()
    for width in range(20, 41):
        result = r._create_box(text='foo', width=width)
        result = result.splitlines()
        assert '+' in result[0]
        assert '-' in result[0]
        assert '+' in result[-1]
        assert '-' in result[-1]
        assert 'foo' in result[1]


def test_create_generic_report():
    r = Report()
    for width in range(20, 41):
        input_queue = Queue()
        input_queue.put(('foo1', 'foobar1'))
        input_queue.put(('foo2', 'foobar2'))
        input_queue.put(('EOF', 'EOF'))

        result = r.create_report_text(input_queue, timeout=30, width=width)
        log.debug("result is:\n{0}".format(result))
        result = result.splitlines()
        assert '-' in result[0]
        assert len(result[0]) == width
        assert 'Report' in result[1]
        assert len(result[1]) == width
        assert '-' in result[2]
        assert 'Plugin: foo1' in result[3]
        assert '-' in result[4]
        assert 'foobar1' in result[5]
        assert '-' in result[6]
        assert 'Plugin: foo2' in result[7]
        assert '-' in result[8]
        assert 'foobar2' in result[9]
        assert '-' in result[10]
        assert len(result) == 11


def test_truncation():
    r = Report()
    for width in range(5, 50):
        for height in range(5, 20):
            log.debug(f"Width: {width}. Height: {height}")
            line = "1" * 30
            for y in range(10):
                text = "\n".join([line, line])
            truncated_line = r._truncate(text=text, max_width=width, max_height=height)
            truncated_line_split = truncated_line.splitlines()
            truncated_line_split = [x for x in truncated_line_split if 'Truncated' not in x]  # Remove the this line was Truncated messages from the results.
            for line in truncated_line_split:
                assert len(line) <= width
            assert len(truncated_line_split) <= height



import os
import sys
import logging
import inspect

import fossor
from fossor.engine import Fossor


logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)


def test_no_popen_usage():
    f = Fossor()
    for plugin in f.list_plugins():
        for line in inspect.getsourcelines(plugin)[0]:
            print(line)
            if 'popen' in line:
                log.error("os.popen is deprecated as of python 2.6, please use the plugin.py shell_call method instead.")
                log.error('os.popen usage located in {plugin} for line {line}'.format(plugin=plugin, line=line))
            assert 'popen' not in line


def test_has_license():
    fossor_path = fossor.__path__[0]
    test_path = fossor_path + '/../test'
    paths = [fossor_path, test_path]
    for path in paths:
        for root, dirs, files in os.walk(path):
            for file in files:
                if file == '__init__.py':
                    continue
                if os.path.splitext(file)[-1] != '.py':
                    continue
                filepath = os.path.join(root, file)

                has_license = False
                license = 'BSD-2 Clause license'
                with open(filepath, 'rt') as f:
                    for line in f:
                        if license in line:
                            has_license = True
                            break
                if not has_license:
                    raise Exception(f"File: {filepath} does not have the license: {license}")



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import psutil
from fossor.variables.variable import Variable


class LogFiles(Variable):
    def run(self, variables):
        if 'Pid' not in variables:
            return
        pid = int(variables['Pid'])
        p = psutil.Process(pid)
        log_files = None

        try:
            log_files = [x.path for x in p.open_files() if x.path.lower().endswith('.log')]
        except psutil.AccessDenied:
            self.log.warn(f"Did not have permission to retrieve open files for process {pid}")

        if log_files:
            return log_files



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import math
import logging
import statistics


log = logging.getLogger(__name__)

# Percent of values with X sigmas, used in various standard 3 sigma checks.
ONE_SIGMA_PERCENT = 0.68
TWO_SIGMA_PERCENT = .9545
THREE_SIGMA_PERCENT = .9973


def within_stdev_percent(values, x_stdev, percent_threshold, min_values=100):
    '''Return True if percent_threshold of values are within x_stdev of the mean.'''
    if len(values) < min_values:
        return True

    mean = statistics.mean(values)
    stdev = statistics.stdev(values)
    found = []
    for v in values:
        diff = abs(mean - v)
        if diff <= (stdev * x_stdev):
            found.append(v)
    percent_found = len(found) / len(values)
    result = percent_found > percent_threshold
    log.debug(f"Within {x_stdev} sigma check was {result}. {percent_found:.2f}%/{percent_threshold:.2f}% within stdev*{x_stdev}. "
              f"Mean: {mean:.2f}. Stdev: {stdev:.2f}. Acceptable range was: {mean - stdev * x_stdev:.2f} - {mean + stdev * x_stdev:.2f}")
    return result


def within_one_sigma(values, percent_threshold=ONE_SIGMA_PERCENT, min_values=100):
    return within_stdev_percent(values=values, x_stdev=1, percent_threshold=percent_threshold, min_values=min_values)


def within_two_sigma(values, percent_threshold=TWO_SIGMA_PERCENT, min_values=100):
    return within_stdev_percent(values, x_stdev=2, percent_threshold=percent_threshold, min_values=min_values)


def within_three_sigma(values, percent_threshold=THREE_SIGMA_PERCENT, min_values=100):
    return within_stdev_percent(values, x_stdev=3, percent_threshold=percent_threshold, min_values=min_values)


def within_all_three_sigma(values, min_values=100):
    '''
    Return false if something does not pass a standard three sigma test.
    If the number of values is less than min_values, will always return False
    '''
    return within_one_sigma(values) and within_two_sigma(values) and within_three_sigma(values)


def abnormal_distribution(values, ignore_zero=False, probability=1e-30):
    '''
    Return True if too many values fall within the same sigma consecutively, meaning the distribution is not normal according to the 3 sigma rule.
    Probability is loosely the likelihood of this test triggering during a normal distribution. A lower value means fewer false positives.
    '''

    result = False

    if ignore_zero:
        values = [value for value in values if value != 0]

    one_sigma_threshold = math.log10(probability) / math.log10(1 - ONE_SIGMA_PERCENT)
    two_sigma_threshold = math.log10(probability) / math.log10(1 - TWO_SIGMA_PERCENT)
    three_sigma_threshold = math.log10(probability) / math.log10(1 - THREE_SIGMA_PERCENT)

    mean = statistics.mean(values)
    stdev = statistics.stdev(values)

    one_sigma = stdev * 1
    two_sigma = stdev * 2
    three_sigma = stdev * 3

    one_sigma_count = 0
    two_sigma_count = 0
    three_sigma_count = 0

    one_sigma_count_max = 0
    two_sigma_count_max = 0
    three_sigma_count_max = 0
    for v in values:
        diff = abs(mean - v)
        if diff < one_sigma:
            one_sigma_count += 1
            one_sigma_count_max = max(one_sigma_count, one_sigma_count_max)
            two_sigma_count = 0
            three_sigma_count = 0
        elif diff < two_sigma:
            one_sigma_count = 0
            two_sigma_count += 1
            two_sigma_count_max = max(two_sigma_count, two_sigma_count_max)
            three_sigma_count = 0
        elif diff < three_sigma:
            one_sigma_count = 0
            two_sigma_count = 0
            three_sigma_count += 1
            three_sigma_count_max = max(three_sigma_count, three_sigma_count_max)
        if one_sigma_count > one_sigma_threshold or \
           two_sigma_count > two_sigma_threshold or \
           three_sigma_count > three_sigma_threshold:
            result = True

    log.debug(f"Abnormal Distribution: {result}. Max consecutive values within one, two, and three sigma in a row: "
              f"{one_sigma_count_max}/{one_sigma_threshold:.2f}, "
              f"{two_sigma_count_max}/{two_sigma_threshold:.2f}, "
              f"{three_sigma_count_max}/{three_sigma_threshold:.2f}. "
              f"Mean: {mean:.2f}. Stdev: {stdev:.2f}.")
    return result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock
from fossor.variables.other_users import OtherUsers


@mock.patch('fossor.variables.other_users.getuser')
@mock.patch('fossor.plugin.Plugin.shell_call')
def test_other_users(sc_mock, user_mock):
    current_user = 'brodriguez'
    user_mock.return_value = current_user
    c = OtherUsers()
    err = ''
    return_code = 0

    # Test for multiple users logged in
    users1 = [current_user, 'pfry', 'hfarnsworth', 'root', 'app']
    out1 = ' '.join(str(u) for u in users1)
    sc_mock.return_value = out1, err, return_code
    assert c.run({}) == 'hfarnsworth pfry'

    # Test for single user logged in
    users2 = [current_user]
    out2 = ' '.join(str(u) for u in users2)
    sc_mock.return_value = out2, err, return_code
    assert c.run({}) is None

    # Test for single user logged in, multiple sessions
    users3 = [current_user, current_user, current_user]
    out3 = ' '.join(str(u) for u in users3)
    sc_mock.return_value = out3, err, return_code
    assert c.run({}) is None

    # Test for single user and other "not interesting users" logged in
    users4 = [current_user, 'root', 'app']
    out4 = ' '.join(str(u) for u in users4)
    sc_mock.return_value = out4, err, return_code
    assert c.run({}) is None



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Returns a dict object. Currently used in unit tests, but could be helpful when interacting with Fossor as a library.
from fossor.reports.report import Report


class DictObject(Report):
    def run(self, variables, report_input, **kwargs):
        result = {}
        while True:
            name, output = report_input.get()
            if name == 'EOF':
                break
            result[name] = output
        return result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

from fossor.checks.similar_log_errors import SimilarLogErrors

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_process_line():
    sle = SimilarLogErrors()
    line = 'java.util.concurrent.TimeoutException'
    assert sle._process_line(line=line) == line

    line = '2017/10/10 00:00:34.251 ERROR [ModelResolverOperator] [stuff] [ranker-jymbii] [random_treeid] Some message text.'
    result = sle._process_line(line=line)
    assert result == ' '.join(line.split()[2:])  # Two lines should match with the datetime stripped out

    line = '2017/10/10 00:00:34.251 INFO [ModelResolverOperator] [stuff] [ranker-jymbii] [random_treeid] Some message text.'
    result = sle._process_line(line=line)
    assert result is None  # Should be None because INFO was passed in

    line = '        at sun.reflect.GeneratedMethodAccessor67.invoke(Unknown Source) ~[?:?]'
    assert sle._process_line(line=line) is None


def test_grouping_similar_lines():
    sle = SimilarLogErrors()
    common_lines = {}
    common_lines['This is a line that is very distinct.'] = 2
    common_lines['The rain in spain falls mainly in the plains'] = 5
    common_lines['The quick brown fox jumps over the fence and goes to the grocery store.'] = 4

    line = 'The quick RANDOM brown fox jumps TEXT over the fence and goes to the grocery store.'
    result = sle._group_similar_log_lines(line=line, common_lines=common_lines)
    log.debug(f"result: {result}")
    assert result['The quick brown fox jumps over the fence and goes to the grocery store.'] == 5



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

import fossor

from fossor.engine import Fossor

import fossor.variables.hostname
import fossor.checks.buddyinfo

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_plugin_import_by_module():
    f = Fossor()
    f.clear_plugins()
    f.add_plugins(fossor)

    assert fossor.checks.buddyinfo.BuddyInfo in f.check_plugins
    assert fossor.variables.examplevariable.ExampleVariable in f.variable_plugins

    assert 'BuddyInfo' in [p.get_name() for p in f.check_plugins]
    assert 'fossor.checks.buddyinfo.BuddyInfo' in [p.get_full_name() for p in f.check_plugins]


def test_plugin_import_by_path():
    f = Fossor()
    f.clear_plugins()

    f.add_plugins(fossor.__path__[0])

    assert 'BuddyInfo' in [p.get_name() for p in f.check_plugins]
    assert 'fossor.local.checks.buddyinfo.BuddyInfo' in [p.get_full_name() for p in f.check_plugins]



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from difflib import SequenceMatcher
from multiprocessing import Pool

from fossor.checks.check import Check


class SimilarLogErrors(Check):
    '''Return uncommon errors from the logs'''

    # TODO add table formatting to show diffs over time?

    SIMILARITY_RATIO = 0.7
    MAX_COMMON_LINES = 20

    # Logs matching these strings will not be included
    log_name_blacklist = ['_configuration.log',
                          '.out',
                          'gc.log',
                          'oor.log',
                          '_databus.log',
                          '_public_access.log',
                          'jetty_thread_pool_stats.log',
                          'engine_access_log.log',
                          ]

    def run(self, variables):
        result = ''

        if 'LogFiles' not in variables:
            return

        log_files = variables['LogFiles']
        log_files = self._remove_blacklisted_logs(log_files)
        self.log.debug(f"log_files: {log_files}")

        pool = Pool(len(log_files))
        pool_results = pool.map(self.get_error_pattern_counts, log_files)

        for filename, common_lines in pool_results:
            lines = []
            for line, count in sorted(common_lines.items(), reverse=True, key=lambda x: x[1]):  # Sort by count
                lines.append(f"Count: {count}. {line.strip()}")
            if lines:
                result += f"filename: {filename}\n"
                result += '\n'.join(lines) + '\n\n'

        if result:
            return result

    def _remove_blacklisted_logs(self, log_files):
        '''Remove blacklisted log_files'''
        for blacklisted_string in self.log_name_blacklist:
            log_files = [x for x in log_files if blacklisted_string not in x]

        return log_files

    def get_error_pattern_counts(self, filename):
        common_lines = {}
        self.log.debug(f"Processing filename: {filename}")
        f = open(filename, 'rt')
        for line in f.readlines():
            line = self._process_line(line)
            if not line:
                continue

            # Group similar log lines
            common_lines = self._group_similar_log_lines(line=line, common_lines=common_lines)

        self.log.debug(f"Finished processing filename: {filename}")
        return filename, common_lines

    def _process_line(self, line):
        first_character = line[0]

        # Skip certain log lines
        # Skip since this is the lower portion of a stacktrace
        if first_character.isspace():
            return
        # Line did not start with a date
        elif first_character.isalpha() and 'Exception' not in line:
            return
        elif line.startswith('Caused by'):
            return
        # Check for an error level
        elif first_character.isdigit():
            line_split = line.split()
            try:
                line_level = line_split[2]
            except IndexError as e:
                self.log.debug(f"IndexError parsing line: {line}")
                return
            if line_level != 'ERROR':
                return
            line = ' '.join(line_split[2:])  # Remove date portion from line
        # Truncate line
        line = f"{line:.200}"
        return line

    def _group_similar_log_lines(self, line, common_lines):
        highest_similarity_ratio = 0
        most_similar_line = None
        for common_line in common_lines:
            similarity_ratio = SequenceMatcher(a=common_line, b=line).quick_ratio()
            if similarity_ratio > highest_similarity_ratio:
                highest_similarity_ratio = similarity_ratio
                most_similar_line = common_line

        if highest_similarity_ratio > self.SIMILARITY_RATIO:
            common_lines[most_similar_line] += 1
        elif len(common_lines) < self.MAX_COMMON_LINES:
            common_lines[line] = 1
        else:
            common_lines.setdefault('Other Error Lines', 0)
            common_lines['Other Error Lines'] += 1
        return common_lines



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import json
import logging

from fossor.checks.check import Check
from fossor.engine import Fossor


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestCheck1(Check):
    def run(self, variables):
        return 'foo1'


class TestCheck2(Check):
    def run(self, variables):
        return 'foo2'


def test_json_report():
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [TestCheck1, TestCheck2]
    f.variable_plugins = set()
    f.add_variable('verbose', True)

    json_result = f.run(report='Json')
    log.debug(f"json_result: {json_result}")

    object_result = json.loads(json_result)
    log.debug(f"object_result: {object_result}")
    assert 'TestCheck1' in object_result

    new_json_result = json.dumps(object_result)
    log.debug(f"new_json_result: {new_json_result}")

    assert 'TestCheck1' in new_json_result
    assert 'TestCheck1' in json_result

    assert 'TestCheck2' in new_json_result
    assert 'TestCheck2' in json_result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import statistics

from asciietch.graph import Grapher

from fossor.checks.check import Check
from fossor.utils.filetools import FileTools
import fossor.utils.anomaly_detection


class SystemLogVolume(Check):

    def __init__(self):
        self.syslog_format = '%b  %d %H:%M:%S'
        self.log_files = [('/var/log/messages', self.syslog_format),
                          ('/var/log/secure', self.syslog_format),
                          ('/var/log/cron', self.syslog_format),
                          ('/var/log/maillog', self.syslog_format),
                          ('/var/log/spooler', self.syslog_format),
                          ]

    def run(self, variables: dict):
        '''Print back system graphs if log volume has changed significantly.'''
        start_time = variables['start_time'] if 'start_time' in variables else None
        end_time = variables['end_time'] if 'end_time' in variables else None
        max_width = variables['MaxPluginOutputWidth'] if 'MaxPluginOutputWidth' in variables else None

        file_line_counts = self.get_line_counts(start_time, end_time)
        g = Grapher()
        result = ''
        for file_path, timestamped_values in file_line_counts:
            if not timestamped_values:
                continue

            values = [value for timestamp, value in timestamped_values.items()]
            if not fossor.utils.anomaly_detection.abnormal_distribution(values, ignore_zero=True):
                continue
            if statistics.mean(values) < 10:
                continue

            graph = g.asciigraph(values=timestamped_values, max_width=max_width, max_height=7, label=True)
            tmp = f"{file_path}:\n{graph}"
            result = '\n'.join([result, tmp])

        return result

    def get_line_counts(self, start_time, end_time):
        file_line_counts = []
        for file_path, date_format in self.log_files:
            line_counts = self.get_file_line_counts(file_path=file_path, date_format=date_format, start_time=start_time, end_time=end_time)
            file_line_counts.append((file_path, line_counts))
        return file_line_counts

    def get_file_line_counts(self, file_path, date_format, start_time=None, end_time=None):
        line_counts = {}
        ft = FileTools()

        ft.date_format = date_format

        try:
            log_lines = ft.get_logs_in_time_range_with_timestamps(file_paths=[file_path], start_time=start_time, end_time=end_time)
            for timestamp, line in log_lines:
                line_counts[timestamp] = line_counts.setdefault(timestamp, 0) + 1
        except PermissionError:
            self.log.warn(f"Did not have permission to access {file_path}")
        except FileNotFoundError:
            self.log.warn(f"{file_path} not found.")
        return line_counts



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import platform

from datetime import datetime, timedelta
from collections import namedtuple

from fossor.checks.check import Check


class NetIFace(Check):
    """NetIFace polls SARs EDEV (network error) statistics and notifies if
    stat > ERROR_THRESHOLD.

    EDEV: statistics on failures (errors) from the network devices are reported"""

    NIC_STATS = {
        # Total number of bad packets received per second
        'rxerr': 0,
        # Total number of errors that happened per second while transmitting packets
        'txerr': 0,
        # Number of received packets dropped per second because of a lack of space in linux buff.
        'rxdrop': 0,
        # Number of transmitted packets dropped per second because of a lack of space in linux buffers
        'txdrop': 0,
        # Number of carrier-errors (link) that happened per second while transmitting packets.
        'txcarr': 0,
    }

    def _parse_sar(self, stdout):
        """Massage stdout from subprocess.Popen in the Sar namedtuple. Errors worth keeping are defined
        within NetIFace.NIC_STATS

        :param stdout: str of data received from subprocess.Popen."""

        Sar = namedtuple('Sar', 'rxerr, coll, txerr, rxdrop, txdrop, txcarr')
        return Sar(*stdout.split()[2:8])

    def run(self, variables: dict):
        """Check for Network Interface Errors using Sar data"""

        # pull in global variables or configure sane defaults for SAR time window.
        # sar -[se] switches only allow for a 24 hour window. This check will further shrink
        # the allowed window to 12 hours.
        minutes = int(variables.get('minutes', 60))

        if minutes > 43200 or minutes < 0:
            minutes = 60

        start_time = (datetime.now() + timedelta(minutes=-minutes)).strftime('%H:%M:%S')
        end_time = datetime.now().strftime('%H:%M:%S')

        # this check relies on data massaged via SAR and therefore Linux only.
        if platform.system() != 'Linux':
            return

        cmd = f'/usr/bin/sar -n EDEV -s {start_time} -e {end_time}'
        self.log.debug(f'executing cmd={cmd}')

        out, err, return_code = self.shell_call(cmd)
        nic_stats = out

        # Get just averages
        nic_stats = [line for line in nic_stats.splitlines() if line.startswith('Average')]

        if not nic_stats:
            self.log.debug("No nics found.")
            return
        for line in nic_stats:
            sar = self._parse_sar(line)
            result = ''

            nic = line.split()[1]

            for stat in NetIFace.NIC_STATS:
                val = float(getattr(sar, stat))
                if val > NetIFace.NIC_STATS[stat]:
                    result += f'stat={stat}, interface={nic} has surpassed threshold. value={val}\n'
            return result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Cli Report
from fossor.reports.report import Report


class StdOut(Report):
    def run(self, variables, report_input, **kwargs):
        if 'TerminalWidth' in variables:
            kwargs['width'] = int(variables['TerminalWidth'])
        timeout = variables['timeout'] if 'timeout' in variables else 600
        truncate = variables.get('truncate', True)
        if truncate:
            kwargs['max_lines_per_plugin'] = 20
        debug = variables['debug'] if 'debug' in variables else False

        # If in debug mode, don't print line by line, print all at once in the print statement below instead
        # This prevents logs lines becoming interspersed with the report.
        stdout = not debug

        result = self.create_report_text(report_input, timeout=timeout, stdout=stdout, truncate=truncate, **kwargs)
        if debug:
            print(result)  # Print the report now that the massive amount of logging has finished
        return result



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
import re

from fossor.checks.check import Check


class RaidStatus(Check):

    def run(self, variables: dict):
        '''Check if there are any drives down in a Raid Array.'''
        mdstat_location = '/proc/mdstat'
        if not os.path.exists(mdstat_location):
            self.log.debug(f"{mdstat_location} does not exist.")
            return
        with open(mdstat_location, 'rt') as f:
            mdstat = f.read()

        arrays = re.findall('\[[U_]+\]', mdstat)
        for array in arrays:
            if '_' in array:
                message = f"Drives down in Raid Array ({array})\n"
                message += f"{mdstat_location} output:\n"
                message += f"{mdstat}"
                return message



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check
import humanfriendly


class BuddyInfo(Check):
    '''Checks for memory fragmentation using the /proc/buddyinfo file'''
    def __init__(self):
        self.zero_threshold = 9

    def run(self, variables):
        out, err, return_code = self.shell_call('cat /proc/buddyinfo')
        self.buddy_info_raw = out
        self.buddy_info_lines = [line for line in self.buddy_info_raw.split('\n') if 'Normal' in line]
        result = ''
        if self.fragmentation():
            result += "Possible memory fragmentation\n"
        else:
            result += "No significant memory fragmentation\n"
        result += f"Smallest Page Size for system is {self._getpagesize()}\n" \
                  f"Therefore, each columns pagesize is: {self._get_column_sizes_human_readable()}\n"
        result += "Output of 'cat /proc/buddyinfo' is:\n" + self.buddy_info_raw
        return result

    def _get_column_sizes_human_readable(self):
        column_count = self._get_columns_len()
        page_size = self._getpagesize()
        column_sizes = []
        for c in range(0, column_count):
            size = humanfriendly.format_size((2**c) * page_size, binary=True)
            column_sizes.append(size)

        return ' '.join(column_sizes)

    def _getpagesize(self):
        '''Return int of the page size for OS'''
        out, err, return_code = self.shell_call('getconf PAGE_SIZE')
        return int(out.strip())

    def _get_columns_len(self):
        '''Get number of columns for first row'''
        for line in self.buddy_info_lines:
            values_str = line.split('Normal')[1]
            columns = values_str.split()
            return len(columns)

    def should_notify(self):
        return self.fragmentation()

    def fragmentation(self):
        for line in self.buddy_info_lines:
            # line example 'Node 0, zone   Normal   1372  95320  85260  28831  45988  38962  24612  15433   8342     85   2723 '
            values_str = line.split('Normal')[1]
            values = [int(value) for value in values_str.split()]
            zero_values = [value for value in values if value < 1]
            if len(zero_values) >= self.zero_threshold:
                return True
        return False


if __name__ == "__main__":
    bi = BuddyInfo()
    print(bi.run(variables={}))



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# Template variable class
from abc import abstractmethod
from fossor.plugin import Plugin


class Variable(Plugin):

    @abstractmethod
    def run(self, variables):
        '''variable should be a dict of variables with values'''
        pass



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check


class DiskUsage(Check):
    """DiskUsage ensures that disk utilization percentage for any mounted partitions
    remains under PERCENTAGE_ALERT ."""

    PERCENTAGE_ALERT = 98

    def run(self, variables):

        out, err, return_code = self.shell_call('df')
        disk_free = out.splitlines()[1:]
        exceeded = []
        for row in disk_free:
            mount = row.split()
            partition = mount[0]
            percentage = int(mount[4].replace('%', ''))

            if percentage >= DiskUsage.PERCENTAGE_ALERT:
                exceeded.append('partition={0} at utilization={1}%'.format(partition, percentage))

        if exceeded:
            return 'Disk utilization is at critical state (> {0}). {1}'.format(DiskUsage.PERCENTAGE_ALERT, ','.join(exceeded))


if __name__ == '__main__':
    c = DiskUsage()
    print(c.run({}))



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check


class ListVariables(Check):

    def run(self, variables):
        '''If verbose is on, list all variables in use'''
        verbose = variables.get('verbose')
        if not verbose:
            return
        result = [f"{name}={value}" for name, value in variables.items()]
        return "\n".join(result)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import io

from setuptools import setup, find_packages


description = 'A plugin oriented tool for automating the investigation of broken hosts and services.'
try:
    with io.open('README.md', encoding="utf-8") as fh:
            long_description = fh.read()
except IOError:
    long_description = description

setup(
    name='fossor',
    version='1.0.0',
    description=description,
    long_description=description,
    url='https://github.com/linkedin/fossor',
    author='Steven R. Callister',
    author_email='scallist@linkedin.com',
    license='License :: OSI Approved :: BSD License',
    packages=find_packages(),
    install_requires=[
        'asciietch==1.0.1',
        'click==6.7',
        'psutil==5.4.1',
        'setproctitle==1.1.10',
        'requests==2.18.4',
        'humanfriendly==4.4.1',
        'parsedatetime==2.4',
        'PTable==0.9.2',
    ],
    tests_require=[
        'pytest>=3.0.6',
        'flake8>=3.5.0',
        'mock>=2.0.0'
        'pytest-timeout>=1.2.0',
    ],
    entry_points={
        'console_scripts': [
            'fossor = fossor.cli:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: POSIX :: Linux'
    ]
)



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
import psutil

from fossor.variables.variable import Variable
from fossor.utils.misc import psutil_exceptions


class Pid(Variable):
    def run(self, variables):
        if 'Product' not in variables:
            return
        product = variables['Product']
        for p in psutil.process_iter():
            try:
                if product.startswith(p.name()):
                    return p.pid
            except psutil_exceptions:
                continue


class PidCwd(Variable):
    def run(self, variables):
        if 'Pid' in variables:
            try:
                return os.readlink('/proc/{pid}/cwd'.format(pid=variables['Pid']))
            except PermissionError:
                pass


class PidExe(Variable):
    def run(self, variables):
        if 'Pid' in variables:
            try:
                return os.readlink('/proc/{pid}/exe'.format(pid=variables['Pid']))
            except PermissionError:
                pass



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
import sys
import logging
import time

from datetime import datetime
from io import StringIO
from fossor.utils.filetools import FileTools
from unittest.mock import patch

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_get_first_last_lines():
    # Using bytes instead of strings since BytesIO I'm using to mock filehandle operations requires this.
    first_line = 'first line\n'
    middle_line = 'middle line\n'
    last_line = 'last line\n'
    data = first_line + middle_line + last_line

    f = StringIO()
    f.write(data)
    orig_pos = f.tell()

    ft = FileTools()
    first, last = ft.get_first_last_lines(f)
    assert orig_pos == f.tell()  # Assert file_handle position returned to where it was originally
    assert first_line == first
    assert last_line == last


def test_get_first_last_lines_empty_line():
    data = ''

    f = StringIO()
    f.write(data)
    orig_pos = f.tell()

    ft = FileTools()
    first, last = ft.get_first_last_lines(f)
    assert orig_pos == f.tell()  # Assert file_handle position returned to where it was originally
    assert first == ''
    assert last == ''


def test_get_first_last_lines_almost_empty():
    data = '\n'

    f = StringIO()
    f.write(data)
    orig_pos = f.tell()

    ft = FileTools()
    first, last = ft.get_first_last_lines(f)
    assert orig_pos == f.tell()  # Assert file_handle position returned to where it was originally
    assert first == '\n'
    assert last == '\n'


def test_get_current_line():
    ft = FileTools()
    text = '\ntest1\ntest2\n'

    log.debug(f"\ntext: {repr(text)}")
    handle = StringIO(text)

    handle.seek(0, os.SEEK_SET)
    line = ft._get_current_line(handle)
    assert line == '\n'
    assert handle.tell() == 0

    # Get 'test1\n' by starting at every single character
    # Each iteration should return 'test1\n'
    for i in range(1, 7):
        handle.seek(i)
        line = ft._get_current_line(handle)
        assert line == 'test1\n'
        assert handle.tell() == 1

    # Test readline then previous line to make sure they return the same line
    handle.seek(1)
    line = handle.readline()
    assert line == 'test1\n'
    line = ft._get_previous_line(handle)
    assert line == 'test1\n'
    assert handle.tell() == 1

    # Get 'test2\n' by starting at every single character
    for i in range(7, 12):
        handle.seek(i)
        line = ft._get_current_line(handle)
        assert line == 'test2\n'
        assert handle.tell() == 7

    # Test without newline at the beginning
    text = 'test1\ntest2\n'
    log.debug(f"\ntext: {repr(text)}")
    handle = StringIO(text)
    for i in range(0, 6):  # Get 'test1\n' by starting at every single character
        handle.seek(i)
        line = ft._get_current_line(handle)
        assert line == 'test1\n'
        assert handle.tell() == 0


def test_get_previous_line():
    ft = FileTools()
    text = ''
    now = time.time()
    for i in range(50):
        dt = datetime.fromtimestamp(now + i)
        dt_str = dt.strftime(ft.date_format)
        # Adding snowman unicode character to test to ensure this doesn't break by reading the middle of a unicode character
        text += f"{dt_str} INFO Snowman-unicode-character:☃ message #{i}\n"

    log.debug(f"\ntext: {repr(text)}")
    handle = StringIO(text)
    handle.seek(0, os.SEEK_END)  # Seek to the last character in the file
    for i in range(49, -1, -1):  # Print messages in reverse order from 49 down to 0. Middle -1 is exclusive, so means 0 in this case.
        line = ft._get_previous_line(handle)
        assert f'message #{i}' in line


def test_gettimestamp():
    ft = FileTools()
    text = ''
    now = int(time.time())
    for i in range(50):
        dt = datetime.fromtimestamp(now + i)
        dt_str = dt.strftime(ft.date_format)
        if i == 19:
            dt_str = '2017/MalformedDate'
        text += f"{dt_str} INFO message #{i}\n"

    log.debug(f"text: {text}")
    handle = StringIO(text)
    for i in range(50):
        if i == 19:
            assert ft._gettimestamp(line=handle.readline()) is None
        else:
            assert now + i == ft._gettimestamp(line=handle.readline())
    assert ft.date_length is not None


def test_binary_search():
    ft = FileTools()
    text = ''
    now = int(time.time())
    for i in range(50):
        dt = datetime.fromtimestamp(now + i)
        dt_str = dt.strftime(ft.date_format)
        text += f"{dt_str} INFO message #{i}\n"

    log.debug(f"text: {text}")
    handle = StringIO(text)

    log.debug("Testing 20/30")
    start_time = now + 20
    end_time = now + 30
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    handle.seek(pos)
    line = handle.readline()
    log.debug(f"20/30 pos: {pos}. line: {line}")
    assert 'message #20' in line

    log.debug("Testing 0/10")
    start_time = now
    end_time = now + 10
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    handle.seek(pos)
    line = handle.readline()
    log.debug(f" 0/10 pos: {pos}. line: {line}")
    assert 'message #0' in line

    log.debug("Testing 45/47")
    start_time = now + 45
    end_time = now + 47
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    handle.seek(pos)
    line = handle.readline()
    log.debug(f" 45/47 pos: {pos}. line: {line}")
    assert 'message #45' in line

    log.debug("Testing -10/-5")
    start_time = now - 10
    end_time = now - 5
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    log.debug(f" -10/-5 pos: {pos}")
    assert pos is None

    log.debug("Testing 55/60")
    start_time = now + 55
    end_time = now + 60
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    log.debug(f" 55/60 pos: {pos}")
    assert pos is None


def test_binary_search_malformed():
    ft = FileTools()
    text = ''
    now = int(time.time())
    for i in range(50):
        dt = datetime.fromtimestamp(now + i)
        dt_str = dt.strftime(ft.date_format)
        if i in range(0, 5) or i in range(15, 20) or i in range(45, 50):
            dt_str = '2017/MalformedDate'
        text += f"{dt_str} INFO message #{i}\n"

    log.debug(f"text: {text}")
    handle = StringIO(text)

    log.debug("Testing 20/30")
    start_time = now + 20
    end_time = now + 30
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    handle.seek(pos)
    line = handle.readline()
    log.debug(f"20/30 pos: {pos}. line: {line}")
    assert 'message #20' in line

    log.debug("Testing 0/10")
    start_time = now
    end_time = now + 10
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    handle.seek(pos)
    line = handle.readline()
    log.debug(f" 0/10 pos: {pos}. line: {line}")
    assert 'message #5' in line

    log.debug("Testing 45/47")
    start_time = now + 45
    end_time = now + 47
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    log.debug(f" 45/47 pos: {pos}.")
    assert pos is None

    log.debug("Testing -10/-5")
    start_time = now - 10
    end_time = now - 5
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    log.debug(f" -10/-5 pos: {pos}")
    assert pos is None

    log.debug("Testing 55/60")
    start_time = now + 55
    end_time = now + 60
    pos = ft._get_first_log_line_position_binary_search(file_handle=handle, start_time=start_time, end_time=end_time)
    log.debug(f" 55/60 pos: {pos}")
    assert pos is None


def test_get_logs_in_time_range():
    ft = FileTools()
    now = int(time.time())

    def _generate_handle_text(start, end):
        handle_text = ''
        for i in range(start, end):
            dt = datetime.fromtimestamp(now + i)
            dt_str = dt.strftime(ft.date_format)
            handle_text += f"{dt_str} INFO message #{i}\n"
        return handle_text

    handle1_text = _generate_handle_text(0, 5)
    handle2_text = _generate_handle_text(10, 15)
    handle3_text = _generate_handle_text(20, 25)

    with patch('fossor.utils.filetools.FileTools._open_log_file') as mo:
        def _mocked_open(*args, **kwargs):
            '''Return mocked handle text if opening specific file_paths, otherwise use the original builtin.open'''
            text = None
            if 'handle1' in args:
                text = handle1_text
            elif 'handle2' in args:
                text = handle2_text
            elif 'handle3' in args:
                text = handle3_text

            if text:
                f = StringIO()
                f.write(text)
                f.seek(0, os.SEEK_SET)
                return f
        mo.side_effect = _mocked_open
        log_generator = ft.get_logs_in_time_range(file_paths=['handle1', 'handle2', 'handle3'])
        lines = [line for line in log_generator]
        for line in lines:
            log.debug(f"timestamp: {ft._gettimestamp(line)}. line: {repr(line)}")
        assert 'message #0\n' in lines[0]
        assert 'message #24\n' in lines[-1]

        # Test sorting
        log_generator = ft.get_logs_in_time_range(file_paths=['handle3', 'handle1', 'handle2'], start_time=None, end_time=None)
        lines = [line for line in log_generator]
        assert 'message #0\n' in lines[0]
        assert 'message #24\n' in lines[-1]

        # Test trimming by time ranges
        log_generator = ft.get_logs_in_time_range(file_paths=['handle3', 'handle1', 'handle2'], start_time=now + 11, end_time=now + 13)
        lines = [line for line in log_generator]
        assert 'message #11\n' in lines[0]
        assert 'message #13\n' in lines[-1]

        log_generator = ft.get_logs_in_time_range(file_paths=['handle3', 'handle1', 'handle2'], start_time=now + 3, end_time=now + 13)
        lines = [line for line in log_generator]
        assert 'message #3\n' in lines[0]
        assert 'message #13\n' in lines[-1]

        log_generator = ft.get_logs_in_time_range(file_paths=['handle3', 'handle1', 'handle2'], start_time=now + 13, end_time=now + 23)
        lines = [line for line in log_generator]
        assert 'message #13\n' in lines[0]
        assert 'message #23\n' in lines[-1]



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging
from fossor.checks.listvariables import ListVariables

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


def test_listvariables_no_verbose():
    variables = {}
    lv = ListVariables()
    assert lv.run(variables) is None


def test_listvariables_verbose():
    variables = {}
    variables['verbose'] = True
    lv = ListVariables()
    assert lv.run(variables) == 'verbose=True'


def test_listvariables():
    variables = {}
    variables['verbose'] = True
    variables['foo'] = 'foo'
    variables['foobar'] = False

    lv = ListVariables()
    output = lv.run(variables)
    assert 'verbose=True' in output
    assert 'foo=foo' in output
    assert 'foobar=False' in output



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import psutil
import logging

from time import sleep
from unittest.mock import patch

from fossor.checks.check import Check
from fossor.engine import Fossor


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)

process_exceptions = (ProcessLookupError, PermissionError, psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess)


class LongCheck(Check):
    def run(self, variables):
        self.log.info("Running long check")
        time_to_sleep = 31.4
        out, err, ret = self.shell_call(f'sleep {time_to_sleep}')
        return "This plugin slept for {0} seconds".format(time_to_sleep)


def test_shell_call_properly_kills():
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 5)
    f.add_variable('verbose', 'True')
    result = f.run(report='DictObject')
    log.debug(f"result is: {result}")
    assert 'Timed out' in result['LongCheck']

    # Confirm that all processes were properly killed before Fossor exited
    for p in psutil.process_iter():
        try:
            assert 'sleep 31.4' != ' '.join(p.cmdline())
        except process_exceptions:
            continue


def test_shell_call_is_needed_to_kill():
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 5)
    f.add_variable('verbose', 'True')
    with patch('fossor.engine.Fossor._terminate_process_group'):
        result = f.run(report='DictObject')
    log.debug(f"result is: {result}")

    # Confirm fossor thinks the plugin ran too long
    assert 'Timed out' in result['LongCheck']

    # Confirm that the sleeping process outlasted lipy-fossor since we patched process group termination to do nothing
    sleeping_pids = []
    for p in psutil.process_iter():
        try:
            if 'sleep 31.4' == ' '.join(p.cmdline()):
                sleeping_pids.append(p)
        except process_exceptions:
            continue
    assert len(sleeping_pids) == 1

    # Clean-up pid by killing it and making sure it is gone
    pid = sleeping_pids[0]
    pid.kill()
    sleep(5)
    sleeping_pids = []
    for p in psutil.process_iter():
        try:
            if 'sleep 31.4' == ' '.join(p.cmdline()):
                sleeping_pids.append(p)
        except process_exceptions:
            continue
    assert len(sleeping_pids) == 0



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import os
from fossor.checks.check import Check


class LoadAvg(Check):
    '''this Check will compare the current load average summaries against the count of CPU cores
    in play, and will alert the user if there are more processes waiting'''
    def run(self, variables):
        ON_LINUX = os.path.isdir('/proc')
        if ON_LINUX:
            with open('/proc/loadavg') as f:
                contents = f.read().strip()
                load_summaries = [float(i) for i in contents.split()[:3]]
            with open('/proc/cpuinfo') as f:
                cpu_count = len([c for c in f.read().splitlines() if c.startswith('processor')])
        else:
            uptime, err, return_code = self.shell_call('uptime')
            contents = uptime.strip()
            load_summaries = [float(i.replace(',', '')) for i in contents.split()[-3:]]

            cpu, err, return_code = self.shell_call('sysctl -n hw.ncpu')
            cpu_count = int(cpu.strip())
        # Alert the user if any of the 1, 5, or 15-minute load averages is greater than the processor count to handle them
        if any(c / cpu_count > 1 for c in load_summaries):
            return 'Load average shows processes queued beyond CPU count!\nCPU Count: {0}\nLoad averages: {1}'.format(
                cpu_count,
                ' '.join(str(x) for x in load_summaries)
            )


if __name__ == '__main__':
    l = LoadAvg()
    print(l.run({}))



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import sys
import logging

from time import sleep

from fossor.checks.check import Check
from fossor.engine import Fossor


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.getLogger(__name__)


class LongCheck(Check):
    def run(self, variables):
        self.log.info("Running long check")
        time_to_sleep = 60
        sleep(time_to_sleep)
        return "This plugin slept for {0} seconds".format(time_to_sleep)


class ShortCheck(Check):
    def run(self, variables):
        self.log.info("Running short check")
        time_to_sleep = 0
        return "This plugin slept for {0} seconds".format(time_to_sleep)


class FailCheck(Check):
    def run(self, variables):
        self.log.info("Running fail check")
        raise Exception("Testing plugin failure")


def test_one_fail_both_short_and_long_running_plugin():
    '''Confirm long running check will not show up in report'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck, ShortCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 1)
    f.add_variable('verbose', 'True')
    result = f.run(report='DictObject')

    assert 'Timed out' in result['LongCheck']
    assert 'This plugin slept' in result['ShortCheck']


def test_timeout_exception_and_success_for_plugins():
    '''Confirm long running check will not show up in report'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck, FailCheck, ShortCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 5)
    f.add_variable('verbose', 'True')
    result = f.run(report='DictObject')

    assert 'Timed out' in result['LongCheck']
    assert 'This plugin slept for' in result['ShortCheck']
    assert 'Failed' in result['FailCheck']


def test_long_running_plugin_verbose():
    '''Confirm long running check does show up in report with verbose on'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 1)
    f.add_variable('verbose', 'True')
    result = f.run(report='DictObject')
    log.debug(f"result is: {result}")
    assert 'Timed out' in result['LongCheck']


def test_long_running_plugin():
    '''Confirm long running check will not show up in report'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [LongCheck]
    f.variable_plugins = set()
    f.add_variable('timeout', 1)
    f.add_variable('verbose', False)
    result = f.run(report='DictObject')
    log.debug(f"result is: {result}")
    assert 'LongCheck' not in result.keys()


def test_short_running_plugin():
    '''Confirm short running check will show up in report'''
    f = Fossor()
    f.check_plugins = set()
    f.check_plugins = [ShortCheck]
    f.variable_plugins = set()
    f.add_variable('verbose', True)
    result = f.run(report='DictObject')
    assert 'ShortCheck' in result.keys()



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.checks.check import Check


class OtherUsers(Check):
    """Checks for other users logged into the host box other than the current user."""
    def run(self, variables):
        if 'OtherUsers' in variables:
            msg = 'Other users logged into this box: '
            user_list = ', '.join(str(u) for u in variables['OtherUsers'].split())
            return msg + user_list + '\n'


if __name__ == '__main__':
    t = OtherUsers()
    print(t.run({}))



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from fossor.variables.variable import Variable


class ExampleVariable(Variable):
    def run(self, variables: dict):
        '''If an object is returned from this method, it will be saved as a variable'''
        if 'Debug' not in variables:
            return
        if variables['Debug'] is True:
            self.log.debug("Logging example")
            return "This is an example variable that only returns when Debug is True."



# Copyright 2017 LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import mock

from fossor.checks.memusage import MemUsage


@mock.patch('platform.system')
@mock.patch('fossor.checks.memusage.MemUsage.get_meminfo')
@mock.patch('fossor.checks.memusage.MemUsage.get_time')
@mock.patch('fossor.plugin.Plugin.shell_call')
def test_mem_usage(sc_mock, time_mock, mem_mock, ps_mock):

    line1 = '[11686.040460] flasherav invoked oom-killer: gfp_mask=0x201da, order=0, oom_adj=0, oom_score_adj=0'
    line2 = '[4680581.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0'
    line3 = '[4680582.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0'

    booted = 1503084682.520215
    cutoff = 4670863.29

    MEM_TOTAL = 100
    MEM_IN_USE = 50

    meminfo = {}
    meminfo['MemTotal'] = MEM_TOTAL
    meminfo['MemFree'] = MEM_TOTAL - MEM_IN_USE
    meminfo['Buffers'] = 0
    meminfo['Cached'] = 0

    out = ''
    err = ''
    return_code = 0
    os_name = "Darwin"

    sc_mock.return_value = out, err, return_code
    mem_mock.return_value = meminfo
    ps_mock.return_value = os_name

    c = MemUsage()

    # Memusage within memusage.MemUsage.CRITICAL_TRESH == OK
    assert c.run({}) is None

    os_name = "Linux"
    ps_mock.return_value = os_name
    assert c.run({}) is None

    # Memusage outside of threshold == WARN
    meminfo['MemFree'] = 5
    assert c.run({}) == 'High Memory Use!\nMemTotal: 100 kib\nUsed: 95 kib (95%)'

    # Memusage outside of threshold, with oom-killer present == WARN
    out = '{}\n{}\n{}'.format(line1, line2, line3)
    sc_mock.return_value = out, err, return_code
    time_mock.return_value = booted, cutoff
    assert c.run({}) == (
        'High Memory Use!\nMemTotal: 100 kib\nUsed: 95 kib (95%)'
        '\n\noom-killer present in dmesg | tail!'
        '\nWed, 11 Oct 2017 23:41:04[4680581.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0\n'
        'Wed, 11 Oct 2017 23:41:05[4680582.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0\n'

        )

    # Memusage within threshold, but with oom-killer present == WARN
    meminfo['MemFree'] = MEM_TOTAL - MEM_IN_USE
    assert c.run({}) == (
        'High Memory Use!\nMemTotal: 100 kib\nUsed: 50 kib (50%)'
        '\n\noom-killer present in dmesg | tail!'
        '\nWed, 11 Oct 2017 23:41:04[4680581.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0\n'
        'Wed, 11 Oct 2017 23:41:05[4680582.563150] perl invoked oom-killer: gfp_mask=0x380da, order=0, oom_score_adj=0\n'
        )
